@@ -11,15 +11,16 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
+
 import logging
 import socket
-from enum import Enum, unique
-
 import requests
 import threading
 import queue
-from requests import HTTPError
 
+from urllib.parse import urljoin
+from enum import Enum, unique
+from requests import HTTPError
 from src.testproject.classes import ActionExecutionResponse
 from src.testproject.enums import ExecutionResultType
 from src.testproject.executionresults import OperationResult
@@ -32,7 +33,12 @@ from src.testproject.rest.messages import (
     StepReport,
     CustomTestReport,
 )
-from src.testproject.sdk.exceptions import SdkException
+from src.testproject.sdk.exceptions import (
+    SdkException,
+    AgentConnectException,
+    InvalidTokenException,
+    ObsoleteVersionException,
+)
 from src.testproject.helpers import SocketHelper
 from src.testproject.sdk.internal.session import AgentSession
 
@@ -40,19 +46,19 @@ from src.testproject.sdk.internal.session import AgentSession
 class AgentClient:
     """Client used to communicate with the TestProject Agent process
 
-        Args:
-            token (str): The development token used to communicate with the Agent
-            capabilities (dict): Additional options to be applied to the driver instance
-            reportsettings (ReportSettings): Settings (project name, job name) to be included in the report
+    Args:
+        token (str): The development token used to communicate with the Agent
+        capabilities (dict): Additional options to be applied to the driver instance
+        reportsettings (ReportSettings): Settings (project name, job name) to be included in the report
 
-        Attributes:
-            _remote_address (str): The Agent endpoint
-            _capabilities (dict): Additional options to be applied to the driver instance
-            _agent_session (AgentSession): stores properties of the current agent session
-            _token (str): The development token used to authenticate with the Agent
-            _reportsettings (ReportSettings): Settings (project name, job name) to be included in the report
-            _sock (socket.Socket): Socket used in communicating with the Agent
-            _queue (queue.Queue): queue holding reports to be sent to Agent in separate thread
+    Attributes:
+        _remote_address (str): The Agent endpoint
+        _capabilities (dict): Additional options to be applied to the driver instance
+        _agent_session (AgentSession): stores properties of the current agent session
+        _token (str): The development token used to authenticate with the Agent
+        _reportsettings (ReportSettings): Settings (project name, job name) to be included in the report
+        _sock (socket.Socket): Socket used in communicating with the Agent
+        _queue (queue.Queue): queue holding reports to be sent to Agent in separate thread
     """
 
     REPORTS_QUEUE_TIMEOUT = 10
@@ -66,7 +72,9 @@ class AgentClient:
         self._queue = queue.Queue()
 
         self._running = True
-        self._reporting_thread = threading.Thread(target=self.__report_worker, daemon=True)
+        self._reporting_thread = threading.Thread(
+            target=self.__report_worker, daemon=True
+        )
         self._reporting_thread.start()
 
         if not self.__start_session():
@@ -80,8 +88,8 @@ class AgentClient:
     def __start_session(self) -> bool:
         """Starts a new development session with the Agent
 
-            Returns:
-                bool: True if the session started successfully, False otherwise
+        Returns:
+            bool: True if the session started successfully, False otherwise
         """
         sdk_version = ConfigHelper.get_sdk_version()
 
@@ -96,7 +104,9 @@ class AgentClient:
             start_session_response.capabilities,
         )
 
-        self._sock = SocketHelper.create_connection(self._remote_address, start_session_response.dev_socket_port)
+        self._sock = SocketHelper.create_connection(
+            self._remote_address, start_session_response.dev_socket_port
+        )
 
         logging.info("Development session started...")
         return True
@@ -104,8 +114,8 @@ class AgentClient:
     def _request_session_from_agent(self) -> SessionResponse:
         """Creates and sends a session request object
 
-            Returns:
-                SessionResponse: object containing the response to the session request
+        Returns:
+            SessionResponse: object containing the response to the session request
         """
         session_request = SessionRequest(self._capabilities, self._reportsettings)
 
@@ -113,12 +123,20 @@ class AgentClient:
 
         try:
             response = self.send_request(
-                "POST", f"{self._remote_address}{Endpoint.DevelopmentSession.value}", session_request.to_json(),
+                "POST",
+                urljoin(self._remote_address, Endpoint.DevelopmentSession.value),
+                session_request.to_json(),
             )
         except requests.exceptions.ConnectionError:
-            logging.error(f"Could not start new session on {self._remote_address}. Is your Agent running?")
-            logging.error("You can download the TestProject Agent from https://app.testproject.io/#/agents")
-            raise SdkException(f"Connection error trying to connect to Agent on {self._remote_address}")
+            logging.error(
+                f"Could not start new session on {self._remote_address}. Is your Agent running?"
+            )
+            logging.error(
+                "You can download the TestProject Agent from https://app.testproject.io/#/agents"
+            )
+            raise AgentConnectException(
+                f"Connection error trying to connect to Agent on {self._remote_address}"
+            )
 
         if not response.passed:
             self.__handle_new_session_error(response)
@@ -135,23 +153,27 @@ class AgentClient:
     def send_request(self, method, path, body) -> OperationResult:
         """Sends HTTP request to Agent
 
-            Args:
-                method (str): HTTP method (GET, POST, ...)
-                path (str): Relative API route path
-                body (dict): Request body
+        Args:
+            method (str): HTTP method (GET, POST, ...)
+            path (str): Relative API route path
+            body (dict): Request body
 
-            Returns:
-                OperationResult: contains result of the sent request
+        Returns:
+            OperationResult: contains result of the sent request
         """
         with requests.Session() as session:
             if method == "GET":
                 response = session.get(path, headers={"Authorization": self._token})
             elif method == "POST":
-                response = session.post(path, headers={"Authorization": self._token}, json=body)
+                response = session.post(
+                    path, headers={"Authorization": self._token}, json=body
+                )
             elif method == "DELETE":
                 response = session.delete(path, headers={"Authorization": self._token})
             else:
-                raise SdkException(f"Unsupported HTTP method {method} in send_request()")
+                raise SdkException(
+                    f"Unsupported HTTP method {method} in send_request()"
+                )
 
         try:
             response.raise_for_status()
@@ -165,23 +187,36 @@ class AgentClient:
         except HTTPError as http_error:
             return OperationResult(False, response.status_code, str(http_error), None)
 
-    def send_action_execution_request(self, codeblock_guid: str, body: dict) -> ActionExecutionResponse:
+    def send_action_execution_request(
+        self, codeblock_guid: str, body: dict
+    ) -> ActionExecutionResponse:
         """Sends HTTP request to Agent
 
-            Args:
-                codeblock_guid (str): The codeblock GUID to be executed
-                body (dict): Parameters to be passed to the Agent
+        Args:
+            codeblock_guid (str): The codeblock GUID to be executed
+            body (dict): Parameters to be passed to the Agent
 
-            Returns:
-                ActionExecutionResponse: contains result of the sent execution request
+        Returns:
+            ActionExecutionResponse: contains result of the sent execution request
         """
 
-        response = self.send_request("POST", f"{self._remote_address}{Endpoint.ActionExecution.value}/{codeblock_guid}", body,)
+        response = self.send_request(
+            "POST",
+            urljoin(
+                urljoin(self._remote_address, Endpoint.ActionExecution.value),
+                codeblock_guid,
+            ),
+            body,
+        )
 
         if not response.passed:
             result = ExecutionResultType.Failed
         else:
-            result = ExecutionResultType.Passed if response.data["resultType"] == "Passed" else ExecutionResultType.Failed
+            result = (
+                ExecutionResultType.Passed
+                if response.data["resultType"] == "Passed"
+                else ExecutionResultType.Failed
+            )
 
         result_data = response.data["outputs"] if response.passed else None
 
@@ -193,9 +228,12 @@ class AgentClient:
         Args:
             driver_command_report: object containing the driver command to be reported
         """
-        endpoint = f"{self._remote_address}{Endpoint.ReportDriverCommand.value}"
 
-        queue_item = QueueItem(report_as_json=driver_command_report.to_json(), url=endpoint)
+        queue_item = QueueItem(
+            report_as_json=driver_command_report.to_json(),
+            url=urljoin(self._remote_address, Endpoint.ReportDriverCommand.value),
+            token=self._token,
+        )
 
         self._queue.put(queue_item, block=False)
 
@@ -205,9 +243,12 @@ class AgentClient:
         Args:
             step_report (StepReport): object containing the step to be reported
         """
-        endpoint = f"{self._remote_address}{Endpoint.ReportStep.value}"
 
-        queue_item = QueueItem(report_as_json=step_report.to_json(), url=endpoint)
+        queue_item = QueueItem(
+            report_as_json=step_report.to_json(),
+            url=urljoin(self._remote_address, Endpoint.ReportStep.value),
+            token=self._token,
+        )
 
         self._queue.put(queue_item, block=False)
 
@@ -217,9 +258,12 @@ class AgentClient:
         Args:
             test_report (CustomTestReport): object containing the test to be reported
         """
-        endpoint = f"{self._remote_address}{Endpoint.ReportTest.value}"
 
-        queue_item = QueueItem(report_as_json=test_report.to_json(), url=endpoint)
+        queue_item = QueueItem(
+            report_as_json=test_report.to_json(),
+            url=urljoin(self._remote_address, Endpoint.ReportTest.value),
+            token=self._token,
+        )
 
         self._queue.put(queue_item, block=False)
 
@@ -230,20 +274,28 @@ class AgentClient:
 
         # Send a final, empty, report to the queue to ensure that
         # the 'running' condition is evaluated one last time
-        self._queue.put(QueueItem(report_as_json=None, url=None), block=False)
+        self._queue.put(
+            QueueItem(report_as_json=None, url=None, token=self._token), block=False
+        )
 
         # Wait until all items have been reported or timeout passes
         self._reporting_thread.join(timeout=self.REPORTS_QUEUE_TIMEOUT)
         if self._reporting_thread.is_alive():
             # Thread is still alive, so there are unreported items
-            logging.warning(f"There are {self._queue.qsize()} unreported items in the queue")
+            logging.warning(
+                f"There are {self._queue.qsize()} unreported items in the queue"
+            )
 
         try:
             self._sock.shutdown(socket.SHUT_RDWR)
             self._sock.close()
-            logging.info(f"Connection to Agent at {self._remote_address} closed successfully")
+            logging.info(
+                f"Connection to Agent at {self._remote_address} closed successfully"
+            )
         except socket.error as msg:
-            logging.error(f"Failed to close socket connection to Agent at {self._remote_address}: {msg}")
+            logging.error(
+                f"Failed to close socket connection to Agent at {self._remote_address}: {msg}"
+            )
             pass
 
     @staticmethod
@@ -254,21 +306,24 @@ class AgentClient:
             response (OperationResult): response from the Agent
         """
         if response.status_code == 401:
-            logging.error("Invalid developer token supplied")
+            logging.error(
+                "Failed to initialize a session with the Agent - invalid developer token supplied"
+            )
             logging.error(
                 "Get your developer token from https://app.testproject.io/#/integrations/sdk?lang=Python"
                 " and set it in the TP_DEV_TOKEN environment variable"
             )
-            logging.error(f"Response from Agent: {response.message}")
-            raise SdkException("Invalid developer token supplied")
+            raise InvalidTokenException(response.message)
         elif response.status_code == 406:
-            logging.error(f"This SDK version ({ConfigHelper.get_sdk_version()}) is incompatible with your Agent version.")
-            logging.error(f"Response from Agent: {response.message}")
-            raise SdkException(f"Invalid SDK version {ConfigHelper.get_sdk_version()}")
+            logging.error(
+                f"Failed to initialize a session with the Agent - obsolete SDK version {ConfigHelper.get_sdk_version()}"
+            )
+            raise ObsoleteVersionException(response.message)
         else:
             logging.error("Failed to initialize a session with the Agent")
-            logging.error(f"Response from Agent: {response.message}")
-            raise SdkException("Failed to initialize a session with the Agent")
+            raise AgentConnectException(
+                f"Agent responded with HTTP status {response.status_code}: [{response.message}]"
+            )
 
     def __report_worker(self):
         """Worker method that is polling the queue for items to report"""
@@ -278,27 +333,30 @@ class AgentClient:
             if isinstance(item, QueueItem):
                 item.send()
             else:
-                logging.warning(f"Unknown object of type {type(item)} found on queue, ignoring it..")
+                logging.warning(
+                    f"Unknown object of type {type(item)} found on queue, ignoring it.."
+                )
             self._queue.task_done()
 
 
 class QueueItem:
     """Helper class representing an item to be reported
 
-        Args:
-            report_as_json (dict): JSON payload representing the item to be reported
-            url (str): Agent endpoint the payload should be POSTed to
+    Args:
+        report_as_json (dict): JSON payload representing the item to be reported
+        url (str): Agent endpoint the payload should be POSTed to
+        token (str): Token used to authenticate with the Agent
 
-        Attributes:
-            _report_as_json (dict): JSON payload representing the item to be reported
-            _url (str): Agent endpoint the payload should be POSTed to
-            _token (str): Token used to authenticate with the Agent
+    Attributes:
+        _report_as_json (dict): JSON payload representing the item to be reported
+        _url (str): Agent endpoint the payload should be POSTed to
+        _token (str): Token used to authenticate with the Agent
     """
 
-    def __init__(self, report_as_json: dict, url: str):
+    def __init__(self, report_as_json: dict, url: str, token: str):
         self._report_as_json = report_as_json
         self._url = url
-        self._token = ConfigHelper.get_developer_token()
+        self._token = token
 
     def send(self):
         """Send a report item to the Agent"""
@@ -307,11 +365,17 @@ class QueueItem:
             return
 
         with requests.Session() as session:
-            response = session.post(self._url, headers={"Authorization": self._token}, json=self._report_as_json,)
+            response = session.post(
+                self._url,
+                headers={"Authorization": self._token},
+                json=self._report_as_json,
+            )
             try:
                 response.raise_for_status()
             except HTTPError:
-                logging.error(f"Reporting to TestProject returned an HTTP {response.status_code}")
+                logging.error(
+                    f"Reporting to TestProject returned an HTTP {response.status_code}"
+                )
                 logging.error(f"Response from Agent: {response.text}")
 
 
